/*
 * true's GAT Nametag
 * Rushed SUPERCON 8 EDITION
 * main MCU firmware
 * 2024 true
 *
 * much of this code is copied from true's DC29 Whiskey Pirates badge,
 * which was in turn copied from the original GAT Nametag.
 *
 * The GAT Nametag used an 8-bit 8051 core MCU.
 * DC29 minibadge used a RISC-V core, and code was hastily ported.
 * this is why things like the menu system are so convoluted;
 *
 * the main MCU is responsible for the following:
 * - rendering OLED UI
 * - rendering LED programs
 * - storing fonts
 * - communicating with I2C devices, including:
 *   - oled display, accelerometer
 *   - eeprom, sub MCU, RGBLED controller
 * - implementing USB interface
 * - implementing BLE (maybe)
 *
 */

#include <hw/ch32sub.h>
#include "CH59x_common.h"

#include "comm/i2c.h"

#include "hw/gat/gat_gpio.h"

#include "led/rgbled.h"

#include "misc/accel.h"

#include "ui/menu.h"
#include "ui/oled.h"

#include "global.h"
#include "port_intr.h"
#include "user_config.h"


// global settings
#define OLED_UPDATE_RATE        64      // framerate of OLED; (256*0.75) / OLED_UPDATE_RATE

// flags
#define FLAG_OLED_UPDATE        (1 << 0)
#define FLAG_RGBLED_RUN_PROG    (1 << 1)



const uint8_t vers[] = "241015a";

uint8_t cpu_use = 0;
uint8_t cpu_max = 0;

uint32_t uptime = 0;
uint16_t uptime_hour;
uint8_t  uptime_min;
uint8_t  uptime_sec;

uint32_t idle_time_menu;
uint32_t idle_time_still;
uint8_t  idle_go_sleep;

static volatile uint8_t flags = 0;

static uint8_t st_tick = 0;             // systick loop counter
static uint8_t oled_tick = 0;           // oled framerate counter





void ch59x_xtal_conf()
{
    HSECFG_Current(HSE_RCur_125);
    HSECFG_Capacitance(HSECap_14p);
}

void systick_init()
{

}

void oled_update_done()
{
    int16_t a;
    int8_t rot;

    // reset oled callback, clear screen, and set default pixel mode and font size
    oled.callback = 0;
    if (!(menu->flags & MENU_FLAG_NO_AUTOCLS)) {
        ssd1306_cls(&oled);
    }
    ssd1306fb_set_color(SSD1306_STATE_SET_PIXEL);
    oled.state &= ~SSD1306_STATE_STR_HALFWIDTH;


    // orientation / flipping flags
    rot = accel_get_rotation(&accel);

    if ((rot > (96+4)) || (rot < (32-4))) {
        sysflags &= ~SYS_OLED_ROTATE_X;
    } else if ((rot > (32+4)) && (rot < (96-4))) {
        sysflags |=  SYS_OLED_ROTATE_X;
    }

    if ((rot > (64+4)) && (rot < 124)) {
        sysflags &= ~SYS_OLED_ROTATE_Y;
    } else if ((rot > 4) && (rot < (64-4))) {
        sysflags |=  SYS_OLED_ROTATE_Y;
    }

    if ((rot < 21) || (rot > (64 + 24))) {
        sysflags &= ~SYS_OLED_REVERSE_CHARS;
    } else if (rot > 24 && rot < (64 + 21)){
        sysflags |=  SYS_OLED_REVERSE_CHARS;
    }


#ifdef MENU_TIMEOUT_TO_NAMETAG
    // root menu idle counting
    if (menu == &menu_0) {
        if (!idle_time_menu) {
            idle_time_menu = uptime;
        } else if ((uptime - idle_time_menu) >= MENU_TIMEOUT_TO_NAMETAG) {
            // been at the root menu too long.
            // return to nametag
            menu_stop(0);
            idle_time_menu = 0;
        }
    } else {
        idle_time_menu = 0;
    }
#endif

    // do menu operations
    menu_tick();

    // calculate CPU usage
    // note: we just consider missed ticks a percent.
    // it's actually more like a tenth. fuck it.
    // a = missed;
    // todo: implement this
    a = 0;

    if (uconf.framemod == UCONF_FRAMERATE_HALF) {
        // the above calculation is tied to framerate.
        // so this will compensate for half framerate mode...
        a >>= 1;
    }

    if (a > cpu_max) {
        cpu_max = a;
    }
    cpu_use = a;


    // reset missed interrupt counter
    // missed = 0;
}



int main()
{
    // configure clock
    ch59x_xtal_conf();
    SetSysClock(SYSCLK_FREQ_USEI2C);

    // enable DC-DC converter; brings significant power saving
    PWR_DCDCCfg(ENABLE);

    // get i2c initialized since most stuff will need it
    i2c_init();

    // read the user config
    uconf_load();

    // initialize aux MCU
    ch32sub_init();
    // and enable RGBLED controller hardware pin so the controller can wake up
    ch32sub_rgb_hwen(1);

    // initialize accelerometer
    accel_init();

    // initialize RGBLED controller
    rgbled_init();

    // configure GAT aux GPIOs, get gat ID
    gat_gpio_init();
    // configure GAT i2c slave
    // not yet implemented

    // configure port-based interrupts (used for ch32sub interrupt)
    port_intr_init();

    // note that system clock speed is decreased after every use of I2C.

    // configure system tick
    systick_init();

    while(1) {
        // sleep when we're doing nothing
        __WFI();

        // only care about aux MCU when all other processing is done
        ch32sub_process();

        // send the last oled frame data
        if (flags & FLAG_OLED_UPDATE) {
            flags &= ~FLAG_OLED_UPDATE;

            /*** oled ***/
            if (oled.callback && !(oled.state & SSD1306_STATE_BUSY)) {
                if (ssd1306_cb_get()) {
                    oled.callback();
                }
            }

            // update
            // only update this frame if we're not in the middle of starting up
            if (uptime) {
                // process other tasks
                if (oled.state & SSD1306_STATE_INITIALIZED) {
                    oled.callback = oled_update_done;
                }
                ssd1306_update();
            }
        }

        // render new OLED frame
        if (flags & FLAG_RGBLED_RUN_PROG) {
            flags &= ~FLAG_RGBLED_RUN_PROG;
            rgbled_runprog(st_tick);
        }
    }
}


void SysTick_Handler(void)
{
    st_tick++;

    if (!st_tick) {
        uptime++;
        uptime_hour = (uint16_t)(uptime / 3600);
        uptime_min  = (uint8_t)((uptime /   60) % 60);
        uptime_sec  = (uint8_t)((uptime       ) % 60);
    }

    // render and update RGBLED at 64Hz
    if ((st_tick & 0x3) == 0x3) {
        // make sure a valid program is selected
        if (uconf.ledprog_rgb_idx > (sizeof(rgb_pgm) / sizeof(rgb_pgm[0]))) {
            uconf.ledprog_rgb_idx = 0;
        }

        // send any rendered data now
        rgbled_send();

        // defer rendering
        flags |= FLAG_RGBLED_RUN_PROG;
    }

    // render and update the oled during non-rgbled frames
    else {
        oled_tick++;
        if (oled_tick >= OLED_UPDATE_RATE) {
            oled_tick = 0;
            flags |= FLAG_OLED_UPDATE;
        }
    }

    // read accelerometer data
    if ((st_tick & 0x7) == 0x7) {
        accel_poll();
    }
}
